<?php

namespace expertinvoice\application\models;

use envoqon\db\ActiveRecord;
use envoqon\db\DbCriteria;
use envoqon\db\QueryBuilder;
use envoqon\utils\Date;
use envoqon\utils\FileHelper;
use envoqon\utils\WpUtil;
use envoqon\web\ActiveDataProvider;
use expertinvoice\application\models\Customer;
use expertinvoice\application\utils\Helper;

/**
 * This is the model class for table "{{invoice}}".
 *
 * The followings are the available columns in table '{{invoice}}':
 * @property integer $id
 * @property integer $receiptNumber
 * @property integer $customerId
 * @property string $createdDate
 * @property string $updatedDate
 * @property integer $termsDays
 * @property integer $salesPerson
 * @property string $privateComment
 * @property string $comment
 * @property integer $termsId
 * @property integer $type
 * @property string $amountDue
 * @property string $totalTax
 * @property integer $hasRefund
 * @property float $discount 
 * @property float $totalDiscount
 * @property string $invoicedDate
 * @property integer $quoteNumber
 * @property string $title
 * @property string $dueDate
 * @property integer $status
 *
 * The followings are the available model relations:
 * @property Customer $customer
 * @property InvoiceItem[] $invoiceItems
 * @property Payment[] $payments
 */
class Invoice extends ActiveRecord {

    const TYPE_INVOICE = 1;
    const TYPE_QUOTE = 2;
    const STATUS_UNPAID = 1;
    const STATUS_PAID = 2;
    const STATUS_PARTIALLY_PAID = 3;

    public $customer;
    private $_totalPaid;

    /**
     * @return string the associated database table name
     */
    public function tableName() {
        return '{{expertinvoice_invoice}}';
    }

    public function getTotalPaid() {
        if (!$this->_totalPaid) {
            $paid = QueryBuilder::get()->select('SUM(amount)')->from(Payment::model()->tableName())->eq('invoiceId', $this->id)->queryScalar();
            $refund = QueryBuilder::get()->select('SUM(amount)')->from(Refund::model()->tableName())->eq('invoiceId', $this->id)->queryScalar();
            $this->_totalPaid = $paid - $refund;
        }
        return $this->_totalPaid;
    }

    /**
     * @return array validation rules for model attributes.
     */
    public function rules() {
        // NOTE: you should only define rules for those attributes that
        // will receive user inputs.
        return array(
            array('customerId,invoicedDate,title,dueDate', 'required'),
            array('receiptNumber', 'required', 'condition' => function($model) {
                    return $model->type == Invoice::TYPE_INVOICE;
                }),
            array('receiptNumber', 'isUnique'),
            array('customerId, termsDays, termsId, type, status, hasRefund', 'numerical', 'integerOnly' => true),
            array('amountDue, totalTax,discount,totalDiscount', 'length', 'max' => 15),
            array('privateComment, comment', 'safe'),
            // The following rule is used by search().
            // @todo Please remove those attributes that should not be searched.
            array('id, receiptNumber, customer, customerId, createdDate, termsDays, privateComment, comment, termsId, type, amountDue, totalTax, hasRefund', 'safe', 'on' => 'search'),
        );
    }

    public function isUnique($attribute, $params) {
        if ($this->type == Invoice::TYPE_INVOICE) {
            $validator = \envoqon\validators\Validator::createValidator('unique', $this, 'receiptNumber');
            $validator->validate($this, array('receiptNumber'));
        }
    }

    /**
     * @return Customer            */
    public function getCustomer() {
        return Customer::model()->findByPk($this->customerId);
    }

    /**
     * @return InvoiceItem[]
     */
    public function getInvoiceItems() {
        return InvoiceItem::model()->findAllByAttributes(array('invoiceId' => $this->getPrimaryKey()));
    }

    /**
     * @return Payment[]
     */
    public function getPayments() {
        return Payment::model()->findAllByAttributes(array('invoiceId' => $this->getPrimaryKey()));
    }

    public function getAttributeLabel($attribute) {
        return __('invoice.' . $attribute, 'expert-invoice');
    }

    /**
     * Retrieves a list of models based on the current search/filter conditions.
     *
     * Typical usecase:
     * - Initialize the model fields with values from filter form.
     * - Execute this method to get ActiveDataProvider instance which will filter
     * models according to data in model fields.
     *
     * @return ActiveDataProvider the data provider that can return the models
     * based on the search/filter conditions.
     */
    public function invoiceSearch() {
        $q = isset($_GET['q']) ? $_GET['q'] : null;
        $criteria = new DbCriteria;
        $criteria->select = "t.*,c.name AS customer,SUM(p.amount) AS totalPaid";
        $criteria->join = "LEFT JOIN {{expertinvoice_customer}} c ON c.id=t.customerId "
                . "LEFT JOIN {{expertinvoice_payment}} p ON p.invoiceId=t.id";

        $criteria->compare('t.id', $q, false, 'OR');
        $criteria->compare('receiptNumber', $q, true, 'OR');
        $criteria->compare('title', $q, true, 'OR');
        $criteria->compare('dueDate', $q, true, 'OR');
        $criteria->compare('privateComment', $q, true, 'OR');
        $criteria->compare('comment', $q, true, 'OR');
        $criteria->compare('c.name', $q, true, 'OR');
        $criteria->addCondition('type=' . self::TYPE_INVOICE);
        $criteria->group = 't.id';

        return new ActiveDataProvider($this, array(
            'criteria' => $criteria,
            'sort' => array(
                'defaultOrder' => 'id DESC'
            )
        ));
    }

    /**
     * Retrieves a list of models based on the current search/filter conditions.
     *
     * Typical usecase:
     * - Initialize the model fields with values from filter form.
     * - Execute this method to get ActiveDataProvider instance which will filter
     * models according to data in model fields.
     *
     * @return ActiveDataProvider the data provider that can return the models
     * based on the search/filter conditions.
     */
    public function quoteSearch() {
        $q = isset($_GET['q']) ? $_GET['q'] : null;
        $criteria = new DbCriteria;
        $criteria->select = "t.*,i.name AS customer";
        $criteria->join = "LEFT JOIN {{expertinvoice_customer}} i ON i.id=t.customerId";

        $criteria->compare('t.id', $q, false, 'OR');
        $criteria->compare('title', $q, true, 'OR');
        $criteria->compare('dueDate', $q, true, 'OR');
        $criteria->compare('receiptNumber', $q, true, 'OR');
        $criteria->compare('privateComment', $q, true, 'OR');
        $criteria->compare('comment', $q, true, 'OR');
        $criteria->compare('i.name', $q, true, 'OR');
        $criteria->addCondition('type=' . self::TYPE_QUOTE);

        return new ActiveDataProvider($this, array(
            'criteria' => $criteria,
            'sort' => array(
                'defaultOrder' => 'id DESC'
            )
        ));
    }

    /**
     * Returns the static model of the specified AR class.
     * Please note that you should have this exact method in all your CActiveRecord descendants!
     * @param string $className active record class name.
     * @return Invoice the static model class
     */
    public static function model($className = __CLASS__) {
        return parent::model($className);
    }

    protected function beforeSave() {
        if (parent::beforeSave()) {
            if ($this->isNewRecord) {
                $this->createdDate = Date::format(time(), Date::PATTERN_ISO_8601_DATETIME);
                $this->updatedDate = $this->createdDate;
                $this->salesPerson = WpUtil::getUserId();
                $this->status = self::STATUS_UNPAID;
                if ($this->type == self::TYPE_QUOTE) {
                    $this->receiptNumber = 0;
                    $maxNumber = QueryBuilder::get()->select('MAX(quoteNumber)')->from('{{expertinvoice_invoice}}')->queryScalar();
                    $this->quoteNumber = max(array(1000, $maxNumber + 1));
                }
            }
            else {
                $this->updatedDate = Date::format(time(), Date::PATTERN_ISO_8601_DATETIME);
            }
            return true;
        }
        return false;
    }

    protected function afterSave() {
        if ($this->isNewRecord) {
            $this->receiptNumber = QueryBuilder::get()->select('receiptNumber')->from('{{expertinvoice_invoice}}')->eq('id', $this->id)->queryScalar();
        }
    }

    protected function afterDelete() {
        parent::afterDelete();
        $models = $this->getInvoiceItems();
        foreach ($models as $model) {
            $model->isInvoiceItem = ($this->type == self::TYPE_INVOICE);
            $model->delete();
        }
        $items = Payment::model()->findAllByAttributes(array('invoiceId'=> $this->id));
        foreach ($items as $item) {
            $item->delete();
        }
        $items = Refund::model()->findAllByAttributes(array('invoiceId'=> $this->id));
        foreach ($items as $item) {
            $item->delete();
        }
    }

    public function saveItems($stocks) {
        $stock = null;
        foreach ($stocks as $attrs) {
            if (!empty($attrs['id'])) {
                $stock = InvoiceItem::model()->findByPk($attrs['id']);
                if ($stock == null) {
                    continue;
                }
            }
            else {
                $stock = new InvoiceItem();
            }
            $stock->setAttributes($attrs);
            $stock->invoiceId = $this->id;
            $stock->isInvoiceItem = $this->type == self::TYPE_INVOICE;
            $stock->save();
        }
    }

    public function deleteItems($stocks) {
        foreach ($stocks as $attrs) {
            if ($attrs['id']) {
                $stock = InvoiceItem::model()->findByPk($attrs['id']);
                if ($stock) {
                    $stock->isInvoiceItem = $this->type == self::TYPE_INVOICE;
                    $stock->delete();
                }
            }
        }
    }

    public function suggest($keyword, $limit = 10) {
        $keyword = '%' . strtr($keyword, array('%' => '\%', '_' => '\_', '\\' => '\\\\')) . '%';
        $models = $this->findAll(array(
            'condition' => 'receiptNumber LIKE :keyword AND type =' . self::TYPE_INVOICE,
            'select' => 'id,receiptNumber',
            'distinct' => true,
            'limit' => $limit,
            'params' => array(
                ':keyword' => $keyword
            ),
        ));
        $names = array();
        foreach ($models as $tag) {
            $names[] = array('label' => $tag->receiptNumber, 'id' => $tag->id);
        }
        return $names;
    }

    public function getStatusLabel() {
        switch ($this->getStatus()) {
            case self::STATUS_UNPAID:
                return __('invoice.unpaid', 'expert-invoice');
            case self::STATUS_PARTIALLY_PAID:
                return __('invoice.partialpaid', 'expert-invoice');
            case self::STATUS_PAID:
                return __('invoice.paid', 'expert-invoice');
        }
    }

    public function getStatus() {
        return $this->status;
    }

    /**
     * 
     * @param float $amount
     * @param integer $paymentMethod
     * @param string $transactionCode
     */
    public function paymentMade($amount, $paymentMethod, $transactionCode) {
        if (!Payment::model()->exists('transactionCode=:t AND paymentMethod=:p', array(':t' => $transactionCode, ':p' => $paymentMethod))) {
            $model = new Payment();
            $model->invoiceId = $this->id;
            $model->amount = $amount;
            $model->paymentMethod = $paymentMethod;
            $model->transactionCode = $transactionCode;
            $model->datePaid = Date::format(time(), Date::PATTERN_ISO_8601_DATETIME);
            $model->save(false);
            try {
                $customer = Customer::model()->findByPk($this->customerId);
                if ($customer && $customer->email) {
                    $template = Template::model()->findByAttributes(array('post_name' => 'expert-invoice-paid-template'));
                    if ($template) {
                        $this->sendEmail($customer, $template);
                    }
                }
            }
            catch (Exception $ex) {
                
            }
        }
    }

    /**
     * 
     * @param Customer $customer
     * @param Template $template
     */
    public function sendEmail($customer, $template) {
        $attachment = $this->getInvoicePdf();
        $headers = array();
        $headers[] = 'From: ' . Option::getAppOption()->name . ' <' . WpUtil::getBlogEmail() . '>';
        $headers[] = 'Content-Type: text/html; charset=UTF-8';
        wp_mail($customer->email, $template->post_title . " #" . $this->receiptNumber, $this->getEmailContent($template, $customer), $headers, $attachment);
    }

    /**
     * 
     * @param type $template
     * @param Customer $customer
     * @param Invoice $model
     */
    private function getEmailContent($template, $customer) {       
        $url = $this->getViewUrl();  
        return strtr($template->post_content, array(
            '{{client}}' => $customer->name,
            '{{link}}' => '<a href="' . $url. '">' . $url . '</a>'
        ));
    }
    
    public function getViewUrl(){
        return add_query_arg('invoiceid', $this->id, get_permalink(Option::getAppOption()->paymentPage));
    }

    private function getInvoicePdf() {
        require_once(FileHelper::getPluginDirPath() . '/expert-invoice/vendor/autoload.php');
//        use Dompdf\Dompdf;
        $title = ($this->type == Invoice::TYPE_INVOICE ? 'Invoice' . $this->receiptNumber : 'Quote' . $this->quoteNumber);
        $fileName = FileHelper::getTmpDir() . DIRECTORY_SEPARATOR . $title . '.pdf';
        $html = $this->getInvoiceTemplate();

        $options = new \Dompdf\Options();
        $options->set('defaultFont', 'Helvetica');
        $options->set('isRemoteEnabled', true);
        // instantiate and use the dompdf class
        $dompdf = new \Dompdf\Dompdf($options);
        $dompdf->loadHtml($html);
        // (Optional) Setup the paper size and orientation
        $dompdf->setPaper('A4', 'landscape');
        // Render the HTML as PDF
        $dompdf->render();
        // Output the generated PDF to Browser
        $output = $dompdf->output();
        file_put_contents($fileName, $output);
//        \envoqon\App::log($fileName, envoqon\logging\Logger::LEVEL_ERROR);
        return $fileName;
    }

    public function getInvoiceTemplate() {
        $customer = Customer::model()->findByPk($this->customerId);
        $filename = $this->type == self::TYPE_INVOICE ? 'invoice/_template.php' : 'quote/_template.php';
        $file = \ExpertInvoicePlugin::$app->getViewPath() . DIRECTORY_SEPARATOR . $filename;
        return Helper::renderFile($file, array('model' => $this, 'customer' => $customer, 'items' => $this->getInvoiceItems()), true);
    }

    public function fields() {
        return array(
            'id', 'customerId', 'receiptNumber', 'invoicedDate', 'discount', 'comment', 'privateComment', 'totalPaid', 'dueDate', 'title');
    }

    public function getNewInvoiceNumber() {
        $maxNumber = QueryBuilder::get()->select('MAX(receiptNumber)')->from('{{expertinvoice_invoice}}')->queryScalar();
        return max(array(Option::getAppOption()->invoiceStartNumber, $maxNumber + 1));
    }

    public function updateStatus() {
        if ($this->totalPaid < $this->amountDue) {
            if ($this->totalPaid <= 0) {
                $this->status = self::STATUS_UNPAID;
            }
            else {
                $this->status = self::STATUS_PARTIALLY_PAID;
            }
        }
        else {
            $this->status = self::STATUS_PAID;
        }
        $this->update(array('status'));
    }

}
