<?php

namespace expertinvoice\application\models;

use envoqon\db\ActiveRecord;
use envoqon\db\DbCriteria;
use envoqon\db\QueryBuilder;
use envoqon\utils\Date;
use envoqon\web\ActiveDataProvider;
use expertinvoice\application\models\InvoiceItem;
use expertinvoice\application\models\StockTransaction;
use expertinvoice\application\models\Tax;


class Product extends ActiveRecord {

    private $oldQuantity = 0;

    
    public function tableName() {
        return '{{expertinvoice_product}}';
    }

    
    public function rules() {
                        return array(
            array('name,taxId,sku', 'required'),
            array('taxId, isManaged, quantity, warnQuantity', 'numerical', 'integerOnly' => true),
            array('name', 'length', 'max' => 255),
            array('sku', 'length', 'max' => 128),
            array('salePrice', 'length', 'max' => 12),
            array('description', 'safe'),
                                    array('id, name, sku, salePrice, taxId, isManaged, quantity, warnQuantity, description', 'safe', 'on' => 'search'),
        );
    }

    
    public function getItems() {
        return InvoiceItem::model()->findAllByAttributes(array('productId' => $this->getPrimaryKey()));
    }

    
    public function getTax() {
        return Tax::model()->findByPk($this->taxId);
    }

    public function getAttributeLabel($attribute) {
        return __('product.' . $attribute, 'expert-invoice');
    }

    
    public function search() {
        $q = isset($_GET['q']) ? $_GET['q'] : null;
        $criteria = new DbCriteria;

        $criteria->compare('id', $q, false, 'OR');
        $criteria->compare('name', $q, true, 'OR');
        $criteria->compare('sku', $q, true, 'OR');
        $criteria->compare('salePrice', $q, true, 'OR');
        $criteria->compare('taxId', $q, false, 'OR');
        $criteria->compare('isManaged', $q, false, 'OR');
        $criteria->compare('quantity', $q, false, 'OR');
        $criteria->compare('warnQuantity', $q, false, 'OR');
        $criteria->compare('description', $q, true, 'OR');

        return new ActiveDataProvider($this, array(
            'criteria' => $criteria,
        ));
    }

    
    public static function model($className = __CLASS__) {
        return parent::model($className);
    }

    protected function beforeSave() {
        if (parent::beforeSave()) {
            if ($this->isNewRecord) {
                $this->createdDate = Date::format(time(), Date::PATTERN_ISO_8601_DATETIME);
                $this->updatedDate = $this->createdDate;
            }
            else {
                $this->updatedDate = Date::format(time(), Date::PATTERN_ISO_8601_DATETIME);
            }
            return true;
        }
        return false;
    }

    protected function afterFind() {
        parent::afterFind();
        $this->oldQuantity = $this->quantity;
    }

    protected function afterSave() {
        parent::afterSave();
        if ($this->isManaged) {
            if (!$this->isNewRecord && ($this->quantity == $this->oldQuantity)) {
                return;
            }
            $transaction = new StockTransaction();
            $transaction->productId = $this->id;
            if ($this->isNewRecord) {
                $transaction->quantity = $this->quantity;
            }
            else {
                $transaction->quantity = $this->quantity - $this->oldQuantity;
            }
            $transaction->save(false);
        }
    }

    public function adjustStock($qty) {
        if ($this->isManaged) {
            $this->quantity += $qty;
            $this->save(false);
        }
    }

    public function suggest($keyword, $limit = 10) {
        $keyword = '%' . strtr($keyword, array('%' => '\%', '_' => '\_', '\\' => '\\\\')) . '%';
        $models = QueryBuilder::get()
                ->select('t.id,t.name,t.description,t.salePrice,t.sku,x.rate AS taxRate,x.id As taxId')
                ->distinct()
                ->from('{{expertinvoice_product}}', 't')
                ->leftJoin('{{expertinvoice_tax}} x', 'x.id=t.taxId')
                ->andWhere('t.name LIKE :keyword OR t.sku LIKE :keyword2', array(
                    ':keyword' => $keyword,
                    ':keyword2' => $keyword,
                ))
                ->limit($limit)
                ->order('t.name')
                ->queryAll();
        $names = array();
        foreach ($models as $tag) {
            $names[] = array('label' => $tag['name'], 'value' => $tag['name'], 'id' => $tag['id'], 'sku' => $tag['sku'],
                'description' => $tag['description'], 'salePrice' => $tag['salePrice'], 'taxRate' => $tag['taxRate'],'taxId' => $tag['taxId']);
        }
        return $names;
    }

    public function fields() {
        return array(
            'id',
            'name',
            'isManaged',
            'description',
            'sku',
            'unitCost',
            'salePrice',
            'taxId',
            'quantity',
            'warnQuantity'
        );
    }

}
