<?php
namespace envoqon\db;

use envoqon\App;
use envoqon\logging\Logger;
use Exception;
use PDO;
use PDOException;
use PDOStatement;


class QueryBuilder {

    const PARAM_PREFIX = ':ycp';

    
    public static $paramCount = 0;
    public $params = array();
    private $unionDistinct = true;
    public $_connection;
    private $_text;
    public $rawSql = false;
    private $query = array();
    private $_paramLog = array();
    private $_statement;
    private $_fetchMode = array(PDO::FETCH_ASSOC);
    public $subQueryAlias = 'tmp';

    private function __construct() {
        
    }

    
    public function __wakeup() {
        $map = array();
        $params = array();
        foreach ($this->params as $name => $value) {
            $newName = self::PARAM_PREFIX . self::$paramCount++;
            $map[$name] = $newName;
            $params[$newName] = $value;
        }
        $this->query['where'] = strtr($this->query['where'], $map);
        $this->params = $params;
    }

    
    public function __sleep() {
        $this->_statement = null;
        return array_keys(get_object_vars($this));
    }

    public function reset() {
        $this->_text = null;
        $this->query = null;
        $this->_statement = null;
        $this->rawSql = false;
        $this->params = array();
        $this->_paramLog = array();
        return $this;
    }

    
    public static function get($db = null) {
        $builder = new QueryBuilder();
        $builder->_connection = $db;
        if ($builder->_connection == null) {
            $builder->_connection = App::getDb();
        }
        return $builder;
    }

    
    public function getConnection() {
        return $this->_connection;
    }

    public function setSubQueryAlias($alias) {
        $this->subQueryAlias = $alias;
        return $this;
    }

    
    public function bind($name, $value) {
        $this->params[$name] = $value;
        return $this;
    }

    
    public function bindValue($name, $value, $dataType = null) {
        $this->prepare();
        if ($dataType === null) {
            $this->_statement->bindValue($name, $value, $this->_connection->getPdoType(gettype($value)));
        }
        else {
            $this->_statement->bindValue($name, $value, $dataType);
        }
        $this->_paramLog[$name] = $value;
        return $this;
    }

    
    public function getQueryString($rebuild = false) {
        if ($this->_text == '' || ($rebuild && !$this->rawSql)) {
            $this->_text = $this->buildQuery();
        }
        if ($this->_connection->tablePrefix !== null && $this->_text != '') {
            $this->_text = preg_replace('/{{(.*?)}}/', $this->_connection->tablePrefix . '\1', $this->_text);
        }
        return $this->_text;
    }

    
    public function setQueryString($query, $params = array()) {
        $this->_text = $query;
        $this->rawSql = true;
        $this->params = array_merge($this->params, $params);
        $this->cancel();
        return $this;
    }

    
    public function eq($column, $value, $enforce = false) {
        $value = "$value";
        if (!$enforce && $value == '') {
            return $this;
        }
        $condition = $this->_connection->quoteColumnName($column) . '=' . self::PARAM_PREFIX . self::$paramCount;
        $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
        $this->addCondition($condition);
        return $this;
    }

    
    public function neq($column, $value, $enforce = false) {
        $value = "$value";
        if (!$enforce && $value == '') {
            return $this;
        }
        $condition = $this->_connection->quoteColumnName($column) . '<>' . self::PARAM_PREFIX . self::$paramCount;
        $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
        $this->addCondition($condition);
        return $this;
    }

    
    public function isNull($column) {
        $this->addCondition($this->_connection->quoteColumnName($column) . ' IS NULL');
        return $this;
    }

    
    public function isNotNull($column) {
        $this->addCondition($this->_connection->quoteColumnName($column) . ' IS NOT NULL');
        return $this;
    }

    
    public function compare($column, $value, $partialMatch = false, $operator = 'AND', $escape = true) {
        if (is_array($value)) {
            if ($value === array())
                return $this;
            return $this->addInCondition($column, $value, $operator);
        }
        else
            $value = "$value";

        if (preg_match('/^(?:\s*(<>|<=|>=|<|>|=))?(.*)$/', $value, $matches)) {
            $value = $matches[2];
            $op = $matches[1];
        }
        else {
            $op = '';
        }

        if ($value === '') {
            return $this;
        }

        if ($partialMatch) {
            if ($op === '') {
                return $this->like($column, $value, $escape, 2, $operator);
            }
            if ($op === '<>') {
                return $this->nl($column, $value, $escape, 2, $operator);
            }
        }
        elseif ($op === '') {
            $op = '=';
        }

        $this->addCondition($column . $op . self::PARAM_PREFIX . self::$paramCount, $operator);
        $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;

        return $this;
    }

    public function compareHaving($column, $value, $partialMatch = false, $escape = true, $operator = 'AND',$enforce = false) {
        if (is_array($value)) {
            if ($value === array()) {
                return $this;
            }
            return $this->in($column, $value);
        }
        else {
            $value = "$value";
        }

        if (preg_match('/^(?:\s*(<>|<=|>=|<|>|=))?(.*)$/', $value, $matches)) {
            $value = $matches[2];
            $op = $matches[1];
        }
        else {
            $op = '';
        }
        if (!$enforce && $value === '') {
            return $this;
        }
        if ($partialMatch) {
            if ($op === '') {
                return $this->like($column, $value, $escape, 2, $operator,$enforce);
            }
            if ($op === '<>') {
                return $this->nlike($column, $value, $escape, 2,$operator, $enforce);
            }
        }
        else if ($op === '') {
            $op = '=';
        }
        $this->having($this->_connection->quoteColumnName($column) . $op . self::PARAM_PREFIX . self::$paramCount);
        $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
        return $this;
    }

    
    public function between($column, $valueStart, $valueEnd) {
        $valueStart = "$valueStart";
        $valueEnd = "$valueEnd";
        if ($valueStart === '' && $valueEnd === '') {
            return $this;
        }
        if (is_string($valueStart) && strpos($valueStart, '.')) {
            $paramStart = $this->_connection->quoteColumnName($valueStart);
        }
        else {
            $paramStart = self::PARAM_PREFIX . self::$paramCount++;
            $this->params[$paramStart] = $valueStart;
        }
        if (is_string($valueEnd) && strpos($valueEnd, '.')) {
            $paramEnd = $this->_connection->quoteColumnName($valueEnd);
        }
        else {
            $paramEnd = self::PARAM_PREFIX . self::$paramCount++;
            $this->params[$paramEnd] = $valueEnd;
        }
        if ($valueEnd && $valueStart) {
            $condition = $this->_connection->quoteColumnName($column) . " BETWEEN $paramStart AND $paramEnd";
        }
        else if ($valueStart) {
            unset($this->params[$paramEnd]);
            $condition = $this->_connection->quoteColumnName($column) . " >= $paramStart";
        }
        else if ($valueEnd) {
            unset($this->params[$paramStart]);
            $condition = $this->_connection->quoteColumnName($column) . " <= $paramEnd";
        }
        $this->addCondition($condition);
        return $this;
    }

    
    public function like($column, $keyword, $escape = true, $wildcard = 2, $operator = 'AND', $enforce = false) {
        if (!$enforce && $keyword == '') {
            return $this;
        }
        if ($escape) {
            $keyword = strtr($keyword, array('%' => '\%', '_' => '\_', '\\' => '\\\\'));
        }
        switch ($wildcard) {
            case 0:
                $keyword = '%' . $keyword;
                break;
            case 1:
                $keyword .= '%';
                break;
            default:
                $keyword = '%' . $keyword . '%';
                break;
        }
        $condition = $this->_connection->quoteColumnName($column) . " LIKE " . self::PARAM_PREFIX . self::$paramCount;
        $this->params[self::PARAM_PREFIX . self::$paramCount++] = $keyword;
        $this->addCondition($condition, $operator);
        return $this;
    }

    
    public function nlike($column, $keyword, $escape = true, $wildcard = 2, $operator = 'AND', $enforce = false) {
        if (!$enforce && $keyword == '') {
            return $this;
        }
        if ($escape) {
            $keyword = strtr($keyword, array('%' => '\%', '_' => '\_', '\\' => '\\\\'));
        }
        switch ($wildcard) {
            case 0:
                $keyword = '%' . $keyword;
                break;
            case 1:
                $keyword .= '%';
                break;
            default:
                $keyword = '%' . $keyword . '%';
                break;
        }
        $condition = $this->_connection->quoteColumnName($column) . " NOT LIKE " . self::PARAM_PREFIX . self::$paramCount;
        $this->params[self::PARAM_PREFIX . self::$paramCount++] = $keyword;
        $this->addCondition($condition, $operator);
        return $this;
    }

    
    public function in($column, $values, $operator = 'AND') {
        if (is_array($values)) {
            if (($n = count($values)) < 1) {
                $this->addCondition('0=1');                 return $this;
            }
            if ($n === 1) {
                $value = reset($values);
                if ($value === null) {
                    $this->addCondition($this->_connection->quoteColumnName($column) . ' IS NULL');
                    return $this;
                }
                $condition = $this->_connection->quoteColumnName($column) . '=' . self::PARAM_PREFIX . self::$paramCount;
                $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
            }
            else {
                $params = array();
                foreach ($values as $value) {
                    $params[] = self::PARAM_PREFIX . self::$paramCount;
                    $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
                }
                $condition = $this->_connection->quoteColumnName($column) . ' IN (' . implode(', ', $params) . ')';
            }
        }
        else {
            $condition = $this->_connection->quoteColumnName($column) . ' IN ( ' . $values->getQueryString() . ' )';
            $this->params = array_merge($this->params, $values->params);
        }
        $this->addCondition($condition, $operator);
        return $this;
    }

    
    public function nin($column, $values) {
        if (is_array($values)) {
            if (($n = count($values)) < 1) {
                $this->addCondition('0=1');                 return $this;
            }
            if ($n === 1) {
                $value = reset($values);
                if ($value === null) {
                    $this->addCondition($this->_connection->quoteColumnName($column) . ' IS NOT NULL');
                    return;
                }
                $condition = $this->_connection->quoteColumnName($column) . '=' . self::PARAM_PREFIX . self::$paramCount;
                $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
            }
            else {
                $params = array();
                foreach ($values as $value) {
                    $params[] = self::PARAM_PREFIX . self::$paramCount;
                    $this->params[self::PARAM_PREFIX . self::$paramCount++] = $value;
                }
                $condition = $this->_connection->quoteColumnName($column) . ' NOT IN (' . implode(', ', $params) . ')';
            }
        }
        else {
            $condition = $this->_connection->quoteColumnName($column) . ' NOT IN ( ' . $values->getQueryString() . ' )';
            $this->params = array_merge($this->params, $values->params);
        }
        $this->addCondition($condition);
        return $this;
    }

    
    public function andWhere($conditions, $params = array()) {
        $this->addCondition($conditions);
        foreach ($params as $name => $value) {
            $this->params[$name] = $value;
        }
        return $this;
    }

    private function addCondition($condition, $operator = 'AND') {
        if (empty($this->query['where'])) {
            $this->query['where'] = $condition;
        }
        else {
            $this->query['where'] = '(' . $this->query['where'] . ') ' . $operator . ' (' . $condition . ')';
        }
        return $this;
    }

    
    public function limit($max, $offset = null) {
        $this->query['limit'] = $max;
        if ($offset !== null) {
            $this->query['offset'] = $offset;
        }
        return $this;
    }

    
    public function offset($offset) {
        $this->query['offset'] = $offset;
        return $this;
    }

    
    public function order($columns, $append = true) {
        if ($append && !empty($this->query['order'])) {
            $this->query['order'] .= ', ' . $columns;
        }
        else {
            $this->query['order'] = $columns;
        }
        return $this;
    }

    
    public function group($columns, $append = true) {
        if ($append && !empty($this->query['group'])) {
            $this->query['group'] .= ', ' . $columns;
        }
        else {
            $this->query['group'] = $columns;
        }
        return $this;
    }

    
    public function select($columns, $append = true) {
        if ($columns instanceof QueryBuilder) {
            $this->params = array_merge($this->params, $columns->params);
            $columns = '( ' . $columns->getQueryString() . ' ) AS ' . $columns->subQueryAlias;
        }
        if ($append && !empty($this->query['select'])) {
            $this->query['select'] .= ', ' . $columns;
        }
        else {
            $this->query['select'] = $columns;
        }
        return $this;
    }

    
    public function from($tables, $alias = '') {
        if (is_string($tables)) {
            $this->query['from'] = $tables;
        }
        else {
            $this->query['from'] = '( ' . $tables->getQueryString() . ' )';
            $this->params = array_merge($this->params, $tables->params);
        }
        $this->query['alias'] = $alias;
        return $this;
    }

    
    public function distinct() {
        $this->query['distinct'] = true;
        return $this;
    }

    
    public function having($condition, $params = array()) {
        if (!empty($this->query['having'])) {
            $this->query['having'] .= ' AND ' . $condition;
        }
        else {
            $this->query['having'] = $condition;
        }
        foreach ($params as $k => $v) {
            $this->params[$k] = $v;
        }
        return $this;
    }

    
    public function union($criteria, $distinct = true) {
        $this->query['union'][] = $criteria->getQueryString();
        $this->params = array_merge($this->params, $criteria->params);
        $this->unionDistinct = $distinct;
        return $this;
    }

    
    public function join($table, $conditions, $params = array()) {
        return $this->joinInternal('join', $table, $conditions, $params);
    }

    
    public function leftJoin($table, $conditions, $params = array()) {
        return $this->joinInternal('left join', $table, $conditions, $params);
    }

    
    public function rightJoin($table, $conditions, $params = array()) {
        return $this->joinInternal('right join', $table, $conditions, $params);
    }

    
    public function crossJoin($table) {
        return $this->joinInternal('cross join', $table);
    }

    
    public function naturalJoin($table) {
        return $this->joinInternal('natural join', $table);
    }

    
    private function joinInternal($type, $table, $conditions = '', $params = array()) {
        if (is_string($table)) {
            if (strpos($table, '(') === false) {
                if (preg_match('/^(.*?)(?i:\s+as\s+|\s+)(.*)$/', $table, $matches)) {                      $table = $this->_connection->quoteTableName($matches[1]) . ' ' . $this->_connection->quoteTableName($matches[2]);
                }
                else {
                    $table = $this->_connection->quoteTableName($table);
                }
            }
        }
        else {
            $this->params = array_merge($this->params, $table->params);
            $table = '( ' . $table->getQueryString() . ' ) AS ' . $table->subQueryAlias;
        }
        if (is_string($conditions)) {
            if ($conditions != '') {
                $conditions = ' ON ' . $conditions;
            }
        }
        else {
            $this->params = array_merge($this->params, $conditions->params);
            $conditions = ' ON ' . $conditions->getCondition();
        }
        $this->query['join'][] = strtoupper($type) . ' ' . $table . $conditions;
        foreach ($params as $name => $value) {
            $this->params[$name] = $value;
        }
        return $this;
    }

    
    public function buildQuery() {
        $sql = !empty($this->query['distinct']) ? 'SELECT DISTINCT ' : 'SELECT ';
        $sql .= !empty($this->query['select']) ? $this->query['select'] : '*';
        if (!empty($this->query['from']))
            $sql .= "\nFROM " . $this->query['from'];
        else
            throw new DbException(App::t('app', 'The DB query must contain the "from" portion.'));
        if (!empty($this->query['alias'])) {
            $sql .= ' AS ' . $this->query['alias'];
        }
        if (!empty($this->query['join'])) {
            $sql .= "\n" . implode("\n", $this->query['join']);
        }
        if (!empty($this->query['where'])) {
            $sql .= "\nWHERE " . $this->query['where'];
        }
        if (!empty($this->query['union'])) {
            if (!$this->unionDistinct) {
                $sql .= "\nUNION ALL " . implode(" UNION ALL  ", $this->query['union']);
            }
            else {
                $sql .= "\nUNION " . implode(" UNION ", $this->query['union']);
            }
        }
        if (!empty($this->query['group']) != '') {
            $sql .= "\nGROUP BY " . $this->query['group'];
        }
        if (!empty($this->query['having'])) {
            $sql .= "\nHAVING " . $this->query['having'];
        }
        if (!empty($this->query['order'])) {
            $sql .= "\nORDER BY " . $this->query['order'];
        }
        $limit = isset($this->query['limit']) ? (int) $this->query['limit'] : -1;
        $offset = isset($this->query['offset']) ? (int) $this->query['offset'] : -1;
        if ($limit >= 0 || $offset > 0) {
            $sql = $this->_connection->getCommandBuilder()->applyLimit($sql, $limit, $offset);
        }
        return $sql;
    }

    
    public function copy() {
        $clone = clone $this;
        $clone->cancel();
        return $clone;
    }

    
    public function getPdoStatement() {
        return $this->_statement;
    }

    
    private function prepare() {
        if ($this->_statement == null) {
            try {
                $this->_statement = $this->getConnection()->getPdoInstance()->prepare($this->getQueryString(true));
                $this->_paramLog = array();
            }
            catch (Exception $e) {
                App::log('Error in preparing SQL: ' . $this->getQueryString(), Logger::LEVEL_ERROR, 'system.db.DbCommand');
                $errorInfo = $e instanceof PDOException ? $e->errorInfo : null;
                throw new DbException(App::t('app', 'DbCommand failed to prepare the SQL statement: {error}', array('{error}' => $e->getMessage())), (int) $e->getCode(), $errorInfo);
            }
        }
    }

    
    public function cancel() {
        $this->_statement = null;
    }

    
    private function bindValues($values) {
        $this->prepare();
        foreach ($values as $name => $value) {
            $this->_statement->bindValue($name, $value, $this->_connection->getPdoType(gettype($value)));
            $this->_paramLog[$name] = $value;
        }
        return $this;
    }

    
    public function execute($params = array()) {
        if ($this->_connection->enableParamLogging && ($pars = array_merge($this->_paramLog, $params)) !== array()) {
            $p = array();
            foreach ($pars as $name => $value)
                $p[$name] = $name . '=' . var_export($value, true);
            $par = '. Bound with ' . implode(', ', $p);
        }
        else
            $par = '';
        App::trace('Executing SQL: ' . $this->getQueryString() . $par, 'system.db.DbCommand');
        try {
            if ($this->_connection->enableProfiling) {
                $_q = $this->interpolateQuery($this->getQueryString(), array_merge($this->_paramLog, $params));
            }
            $this->prepare();
            $this->bindValues($this->params);
            if ($params === array()) {
                $this->_statement->execute();
            }
            else {
                $this->_statement->execute($params);
            }
            $n = $this->_statement->rowCount();
            if ($this->_connection->enableProfiling)
            return $n;
        }
        catch (Exception $e) {
            if ($this->_connection->enableProfiling)
            $errorInfo = $e instanceof PDOException ? $e->errorInfo : null;
            $message = $e->getMessage();
            App::log(App::t('app', 'DbCommand::execute() failed: {error}. The SQL statement executed was: {sql}.', array('{error}' => $message, '{sql}' => $this->getQueryString() . $par)), Logger::LEVEL_ERROR, 'system.db.DbCommand');
            if (ENVOQON_DEBUG)
                $message .= '. The SQL statement executed was: ' . $this->getQueryString() . $par;
            throw new DbException(App::t('app', 'DbCommand failed to execute the SQL statement: {error}', array('{error}' => $message)), (int) $e->getCode(), $errorInfo);
        }
    }

    
    public function query($params = array()) {
        return $this->queryInternal('', 0, $params);
    }

    
    public function queryAll($fetchAssociative = true, $params = array()) {
        return $this->queryInternal('fetchAll', $fetchAssociative ? $this->_fetchMode : PDO::FETCH_NUM, $params);
    }

    
    public function queryRow($fetchAssociative = true, $params = array()) {
        return $this->queryInternal('fetch', $fetchAssociative ? $this->_fetchMode : PDO::FETCH_NUM, $params);
    }

    
    public function queryScalar($params = array()) {
        $result = $this->queryInternal('fetchColumn', 0, $params);
        if (is_resource($result) && get_resource_type($result) === 'stream')
            return stream_get_contents($result);
        else
            return $result;
    }

    
    public function queryColumn($params = array()) {
        return $this->queryInternal('fetchAll', array(PDO::FETCH_COLUMN, 0), $params);
    }

    
    private function queryInternal($method, $mode, $params = array()) {
        $params = array_merge($this->params, $params);
        if ($this->_connection->enableParamLogging && ($pars = array_merge($this->_paramLog, $params)) !== array()) {
            $p = array();
            foreach ($pars as $name => $value)
                $p[$name] = $name . '=' . var_export($value, true);
            $par = '. Bound with ' . implode(', ', $p);
        }
        else
            $par = '';
        App::trace('Querying SQL: ' . $this->getQueryString() . $par, 'system.db.DbCommand');
        if ($this->_connection->queryCachingCount > 0 && $method !== '' && $this->_connection->queryCachingDuration > 0 && $this->_connection->queryCacheID !== false && ($cache = App::app()->getComponent($this->_connection->queryCacheID)) !== null) {
            $this->_connection->queryCachingCount--;
            $cacheKey = 'app:dbquery' . $this->_connection->connectionString . ':' . $this->_connection->username;
            $cacheKey .= ':' . $this->getQueryString() . ':' . serialize(array_merge($this->_paramLog, $params));
            if (($result = $cache->get($cacheKey)) !== false) {
                App::trace('Query result found in cache', 'system.db.DbCommand');
                return $result[0];
            }
        }
        try {
            if ($this->_connection->enableProfiling) {
                $_q = $this->interpolateQuery($this->getQueryString(), array_merge($this->_paramLog, $params));
            }
            $this->prepare();
            $this->bindValues($this->params);
            if ($params === array())
                $this->_statement->execute();
            else
                $this->_statement->execute($params);
            if ($method === '')
                $result = new DbDataReader($this);
            else {
                $mode = (array) $mode;
                call_user_func_array(array($this->_statement, 'setFetchMode'), $mode);
                $result = $this->_statement->$method();
                $this->_statement->closeCursor();
            }
            if ($this->_connection->enableProfiling)
            if (isset($cache, $cacheKey))
                $cache->set($cacheKey, array($result), $this->_connection->queryCachingDuration, $this->_connection->queryCachingDependency);
            return $result;
        }
        catch (Exception $e) {
            if ($this->_connection->enableProfiling)
            $errorInfo = $e instanceof PDOException ? $e->errorInfo : null;
            $message = $e->getMessage();
            App::log(App::t('app', 'DbCommand::{method}() failed: {error}. The SQL statement executed was: {sql}.', array('{method}' => $method, '{error}' => $message, '{sql}' => $this->getQueryString() . $par)), Logger::LEVEL_ERROR, 'system.db.DbCommand');
            if (ENVOQON_DEBUG)
                $message .= '. The SQL statement executed was: ' . $this->getQueryString() . $par;
            throw new DbException(App::t('app', 'DbCommand failed to execute the SQL statement: {error}', array('{error}' => $message)), (int) $e->getCode());
        }
    }

    
    public function setFetchMode($mode) {
        $params = func_get_args();
        $this->_fetchMode = $params;
        return $this;
    }

    private function interpolateQuery($query, $params) {
        $keys = array();
        $values = array();
        foreach ($params as $k => $v) {
            if (is_string($k)) {
                $keys[] = "/$k/";
            }
            else {
                $keys[] = "/[?]/";
            }
            $values[] = '"' . $v . '"';
        }
        return preg_replace($keys, $values, $query, 1);
    }

}
