<?php

namespace envoqon\utils;


class FileHelper {

    
    public static function getExtension($path) {
        return pathinfo($path, PATHINFO_EXTENSION);
    }

    
    public static function copyDirectory($src, $dst, $options = array()) {
        $fileTypes = array();
        $exclude = array();
        $level = -1;
        extract($options);
        if (!is_dir($dst))
            self::createDirectory($dst, isset($options['newDirMode']) ? $options['newDirMode'] : null, true);

        self::copyDirectoryRecursive($src, $dst, '', $fileTypes, $exclude, $level, $options);
    }

    
    public static function removeDirectory($directory, $options = array()) {
        if (!isset($options['traverseSymlinks']))
            $options['traverseSymlinks'] = false;
        $items = glob($directory . DIRECTORY_SEPARATOR . '{,.}*', GLOB_MARK | GLOB_BRACE);
        foreach ($items as $item) {
            if (basename($item) == '.' || basename($item) == '..')
                continue;
            if (substr($item, -1) == DIRECTORY_SEPARATOR) {
                if (!$options['traverseSymlinks'] && is_link(rtrim($item, DIRECTORY_SEPARATOR)))
                    unlink(rtrim($item, DIRECTORY_SEPARATOR));
                else
                    self::removeDirectory($item, $options);
            }
            else
                unlink($item);
        }
        if (is_dir($directory = rtrim($directory, '\\/'))) {
            if (is_link($directory))
                unlink($directory);
            else
                rmdir($directory);
        }
    }

    
    public static function findFiles($dir, $options = array()) {
        $fileTypes = array();
        $exclude = array();
        $level = -1;
        $absolutePaths = true;
        extract($options);
        $list = self::findFilesRecursive($dir, '', $fileTypes, $exclude, $level, $absolutePaths);
        sort($list);
        return $list;
    }

    
    protected static function copyDirectoryRecursive($src, $dst, $base, $fileTypes, $exclude, $level, $options) {
        if (!is_dir($dst))
            self::createDirectory($dst, isset($options['newDirMode']) ? $options['newDirMode'] : null, false);

        $folder = opendir($src);
        if ($folder === false)
            throw new Exception('Unable to open directory: ' . $src);
        while (($file = readdir($folder)) !== false) {
            if ($file === '.' || $file === '..')
                continue;
            $path = $src . DIRECTORY_SEPARATOR . $file;
            $isFile = is_file($path);
            if (self::validatePath($base, $file, $isFile, $fileTypes, $exclude)) {
                if ($isFile) {
                    copy($path, $dst . DIRECTORY_SEPARATOR . $file);
                    if (isset($options['newFileMode']))
                        @chmod($dst . DIRECTORY_SEPARATOR . $file, $options['newFileMode']);
                }
                elseif ($level)
                    self::copyDirectoryRecursive($path, $dst . DIRECTORY_SEPARATOR . $file, $base . '/' . $file, $fileTypes, $exclude, $level - 1, $options);
            }
        }
        closedir($folder);
    }

    
    protected static function findFilesRecursive($dir, $base, $fileTypes, $exclude, $level, $absolutePaths) {
        $list = array();
        $handle = opendir($dir . $base);
        if ($handle === false)
            throw new Exception('Unable to open directory: ' . $dir);
        while (($file = readdir($handle)) !== false) {
            if ($file === '.' || $file === '..')
                continue;
            $path = substr($base . DIRECTORY_SEPARATOR . $file, 1);
            $fullPath = $dir . DIRECTORY_SEPARATOR . $path;
            $isFile = is_file($fullPath);
            if (self::validatePath($base, $file, $isFile, $fileTypes, $exclude)) {
                if ($isFile)
                    $list[] = $absolutePaths ? $fullPath : $path;
                elseif ($level)
                    $list = array_merge($list, self::findFilesRecursive($dir, $base . '/' . $file, $fileTypes, $exclude, $level - 1, $absolutePaths));
            }
        }
        closedir($handle);
        return $list;
    }

    
    protected static function validatePath($base, $file, $isFile, $fileTypes, $exclude) {
        foreach ($exclude as $e) {
            if ($file === $e || strpos($base . '/' . $file, $e) === 0)
                return false;
        }
        if (!$isFile || empty($fileTypes))
            return true;
        if (($type = self::getExtension($file)) !== '')
            return in_array($type, $fileTypes);
        else
            return false;
    }

    
    public static function getMimeType($file, $magicFile = null, $checkExtension = true) {
        if (function_exists('finfo_open')) {
            $options = defined('FILEINFO_MIME_TYPE') ? FILEINFO_MIME_TYPE : FILEINFO_MIME;
            $info = $magicFile === null ? finfo_open($options) : finfo_open($options, $magicFile);

            if ($info && ($result = finfo_file($info, $file)) !== false)
                return $result;
        }

        if (function_exists('mime_content_type') && ($result = mime_content_type($file)) !== false)
            return $result;

        return $checkExtension ? self::getMimeTypeByExtension($file) : null;
    }

    
    public static function getMimeTypeByExtension($file, $magicFile = null) {
        static $extensions, $customExtensions = array();
        if ($magicFile === null && $extensions === null)
            $extensions = require(App::getPathOfAlias('system.utils.mimeTypes') . '.php');
        elseif ($magicFile !== null && !isset($customExtensions[$magicFile]))
            $customExtensions[$magicFile] = require($magicFile);
        if (($ext = self::getExtension($file)) !== '') {
            $ext = strtolower($ext);
            if ($magicFile === null && isset($extensions[$ext]))
                return $extensions[$ext];
            elseif ($magicFile !== null && isset($customExtensions[$magicFile][$ext]))
                return $customExtensions[$magicFile][$ext];
        }
        return null;
    }

    
    public static function getExtensionByMimeType($file, $magicFile = null) {
        static $mimeTypes, $customMimeTypes = array();
        if ($magicFile === null && $mimeTypes === null)
            $mimeTypes = require(App::getPathOfAlias('system.utils.fileExtensions') . '.php');
        elseif ($magicFile !== null && !isset($customMimeTypes[$magicFile]))
            $customMimeTypes[$magicFile] = require($magicFile);
        if (($mime = self::getMimeType($file)) !== null) {
            $mime = strtolower($mime);
            if ($magicFile === null && isset($mimeTypes[$mime]))
                return $mimeTypes[$mime];
            elseif ($magicFile !== null && isset($customMimeTypes[$magicFile][$mime]))
                return $customMimeTypes[$magicFile][$mime];
        }
        return null;
    }

    
    public static function createDirectory($dst, $mode = null, $recursive = false) {
        if ($mode === null)
            $mode = 0777;
        $prevDir = dirname($dst);
        if ($recursive && !is_dir($dst) && !is_dir($prevDir))
            self::createDirectory(dirname($dst), $mode, true);
        $res = mkdir($dst, $mode);
        @chmod($dst, $mode);
        return $res;
    }

    
    public static function deleteFolderRecursive($folder) {
        if (is_dir($folder)) {
            $handle = opendir($folder);
            while (($filename = readdir($handle)) !== false) {
                if ($filename !== '.' && $filename !== '..') {                     $path = $folder . DIRECTORY_SEPARATOR . $filename;
                    if (is_file($path)) {
                        @unlink($path);
                    }
                    else {
                        self::deleteFolderRecursive($path);
                    }
                }
            }
            closedir($handle);             return rmdir($folder);
        }
        return false;
    }

    public static function createWritableFolder($folder) {
        if (!file_exists($folder)) {
            mkdir($folder, 0755, true);
        }
        chmod($folder, 0775);
    }

    
    public static function deleteFiles($folder) {
        if (is_dir($folder)) {
            $handle = opendir($folder);
            while (($filename = readdir($handle)) !== false) {
                if ($filename !== '.' && $filename !== '..') {                     $path = $folder . DIRECTORY_SEPARATOR . $filename;
                    if (is_file($path)) {
                        @unlink($path);
                    }
                    else {
                        self::deleteFolderRecursive($path);
                    }
                }
            }
            closedir($handle);         }
        return false;
    }

    public static function directoryNames($folder) {
        $names = array();
        if (is_dir($folder)) {
            $handle = opendir($folder);
            while (($filename = readdir($handle)) !== false) {
                if ($filename !== '.' && $filename !== '..') {                     $path = $folder . DIRECTORY_SEPARATOR . $filename;
                    if (is_dir($path)) {
                        $names[] = $filename;
                    }
                }
            }
            closedir($handle);         }
        return $names;
    }

    
    public static function bytesToSize($bytes) {
        $units = array('B', 'KB', 'MB', 'GB', 'TB');
        $u = 0;
        while ((round($bytes / 1024) > 0) && ($u < 4)) {
            $bytes = $bytes / 1024;
            $u++;
        }
        return (number_format($bytes, 1, '.', '') . " " . $units[$u]);
    }

    
    public static function download($file, $mimeType = '') {
        if (file_exists($file)) {
            set_time_limit(0);             $baseName = basename($file);
            $mimeType = ($mimeType == '') ? self::getMimeType($file) : $mimeType;
            $length = filesize($file);
                        ob_clean();
            header('Content-Encoding:identity');
                        @ob_end_clean();
                        if (ini_get('zlib.output_compression')) {
                ini_set('zlib.output_compression', 'off');
            }
            header('Content-Description: File Transfer');
            header("Content-Type: {$mimeType}");
            header("Content-Disposition: attachment; filename={$baseName}");
            header('Content-Transfer-Encoding: binary');
            header('Accept-Ranges:bytes');
                        header('Expires: Tue,1 Feb 2011 0:0:0 GMT');
            header('Cache-Control: private');
            header('Pragma: private');
            $range = 0;
                        if (isset($_SERVER['HTTP_RANGE'])) {
                list(, $range) = explode("=", $_SERVER['HTTP_RANGE'], 2);
                list($range) = explode(",", $range, 2);
                list($range, $rangeEnd) = explode("-", $range);
                $range = intval($range);
                if (!$rangeEnd) {
                    $rangeEnd = intval($length - 1);
                }
                else {
                    $rangeEnd = intval($rangeEnd);
                }
                header("HTTP/1.1 206 Partial Content");
                $bytes = $range - $rangeEnd / $length;
                header("Content_Range:bytes $bytes");
                $length = $rangeEnd - $range + 1;
            }
            header("Content-Length: $length");
            $chunkSize = 1000 * 1024;
            $byteSent = 0;
            if (($handle = fopen($file, 'r')) !== false) {
                if (isset($_SERVER['HTTP_RANGE'])) {
                    fseek($handle, $range);
                }
                while (!feof($handle) && !connection_aborted() && $byteSent < $length) {
                    $buffer = fread($handle, $chunkSize);
                    echo $buffer;
                    $byteSent += strlen($buffer);
                }
                flush();
                fclose($handle);
                return true;
            }
        }
        return false;
    }

    private static function encryptChunks($string, $key) {
        $key = pack('H*', $key);
        if (extension_loaded('mcrypt') === true) {
            return mcrypt_encrypt(MCRYPT_BLOWFISH, substr($key, 0, mcrypt_get_key_size(MCRYPT_BLOWFISH, MCRYPT_MODE_ECB)), $string, MCRYPT_MODE_ECB, mcrypt_create_iv(mcrypt_get_iv_size(MCRYPT_BLOWFISH, MCRYPT_MODE_ECB), MCRYPT_RAND));
        }
        return false;
    }

    private static function decryptChunks($string, $key) {
        $key = pack('H*', $key);
        if (extension_loaded('mcrypt') === true) {
            return mcrypt_decrypt(MCRYPT_BLOWFISH, substr($key, 0, mcrypt_get_key_size(MCRYPT_BLOWFISH, MCRYPT_MODE_ECB)), $string, MCRYPT_MODE_ECB, mcrypt_create_iv(mcrypt_get_iv_size(MCRYPT_BLOWFISH, MCRYPT_MODE_ECB), MCRYPT_RAND));
        }
        return false;
    }

    public static function encryptFile($source, $destination, $key) {
        return self::cryptFile($source, $destination, $key, 'encrypt');
    }

    public static function decryptFile($source, $destination, $key) {
        return self::cryptFile($source, $destination, $key, 'decrypt');
    }

    private static function cryptFile($source, $destination, $key, $op) {
        $buffer = '';
        $inHandle = fopen($source, 'rb');
        $outHandle = fopen($destination, 'wb+');
        if ($inHandle === false) {
            return false;
        }
        if ($outHandle === false) {
            return false;
        }
        $key = md5($key);
        while (!feof($inHandle)) {
            $buffer = fread($inHandle, 102400);             if ($op == "encrypt") {
                $buffer = self::encryptChunks($buffer, $key);
            }
            elseif ($op == "decrypt") {
                $buffer = self::decryptChunks($buffer, $key);
            }
            fwrite($outHandle, $buffer);
        }
        fclose($inHandle);
        fclose($outHandle);
        return true;
    }

    public static function cleanFileName($name) {
        return str_replace(array(' ', "(", ")", "#", "@", "%", "^", "&", "+", "=", "!", ">", "<", "?", ",", "*"), '_', $name);
    }

    
    public static function zip($files, $filename) {
        if (file_exists($filename)) {
            @unlink($filename);
        }
        $zip = new ZipArchive();
        if ($zip->open($filename, ZIPARCHIVE::CREATE) !== true) {
            return false;
        }
        foreach ($files as $file) {
            if (is_dir($file)) {
                $iterator = new RecursiveIteratorIterator(new RecursiveDirectoryIterator($file));
                foreach ($iterator as $key => $value) {
                    $zip->addFile(realpath($key), $key);
                }
            }
            else {
                                                if (file_exists($file)) {
                    $zip->addFile($file, basename($file));
                }
            }
        }
        $zip->close();
        return true;
    }

    public static function unzip($filename, $destination) {
        $zip = new ZipArchive();
        if ($zip->open($filename) !== true) {
            return false;
        }
        return $zip->extractTo($destination);
    }
    
    public static function getTmpDir() {
        return get_temp_dir();
    }
    
    public static function getUploadDirPath(){
        $upload_dir = wp_upload_dir();
        return $upload_dir['path'];
    }
    
    public static function getUploadDirUrl(){
        $upload_dir = wp_upload_dir();
        return $upload_dir['url'];
    }
    
    
    public static function getPluginDirPath(){
        return WP_PLUGIN_DIR;
    }
    
    
    public static function getPluginDirUrl(){
        return WP_PLUGIN_URL;
    }

}
