<?php

namespace envoqon\web;

use envoqon\App;
use envoqon\web\HttpException;
use envoqon\collections\CMap;
use envoqon\web\filter\FilterChain;
use envoqon\web\filter\AccessControlFilter;


class Controller extends \envoqon\base\Component {

    
    public $defaultAction = 'index';
    private $_id;
    private $widgets = array();
    private $_pageTitle;
    private $_actionId;
    private $_widgetStack = array();

    
    public function __construct($id) {
        $this->_id = $id;
    }

    
    public function getId() {
        return $this->_id;
    }

    
    public function getActionId() {
        return $this->_actionId;
    }

    
    public function filters() {
        return array();
    }

    
    public function accessRules() {
        return array();
    }

    
    public function run($actionID) {
        if (($action = $this->createAction($actionID)) !== null) {
            $this->_actionId = $actionID;
            if ($this->beforeControllerAction($this, $action)) {
                $this->runActionWithFilters($action, $this->filters());
                $this->afterControllerAction($this, $action);
            }
        }
        else {
            $this->missingAction($actionID);
        }
    }

    
    public function runActionWithFilters($action, $filters) {
        if (empty($filters)) {
            $this->runAction($action);
        }
        else {
            FilterChain::create($this, $action, $filters)->run();
        }
    }

    
    public function runAction($action) {
        if ($action->runWithParams($_GET) === false) {
            $this->invalidActionParams($action);
        }
    }

    
    public function invalidActionParams($action) {
        throw new HttpException(400, App::t('app', 'Your request is invalid.'));
    }

    
    public function getPageTitle() {
        return $this->_pageTitle;
    }

    
    public function setPageTitle($value) {
        $this->_pageTitle = $value;
    }

    
    public function createAction($actionID) {
        if ($actionID === '') {
            $actionID = $this->defaultAction;
        }
        if (method_exists($this, 'action' . $actionID) && strcasecmp($actionID, 's')) {             return new InlineAction($this, $actionID);
        }
        else {
            throw new \Exception(App::t('app', 'Action {class} not implemented', array('{class}' => $actionID)));
        }
    }

    
    public function createUrl($menuPage, $actionId, $params = array()) {
        $params['page'] = $menuPage;
        $params['action'] = $actionId;
        return add_query_arg($params, App::instance()->adminBaseUrl);
    }

    
    public function createAjaxUrl($route, $params = array()) {
        $params['action'] = App::instance()->getAjaxAction();
        $params['r'] = urlencode($route);
        return add_query_arg($params, admin_url('admin-ajax.php'));
    }

    
    public function createFrontPageUrl($route, $params = array()) {        $params['r'] = urlencode($route);
        global $wp;
        return add_query_arg($params, home_url($wp->request));
    }

    
    public function redirect($pageName, $actionId, $params = array()) {
        $params['page'] = $pageName;
        $params['action'] = $actionId;
        $url = add_query_arg($params, App::instance()->adminBaseUrl);
        echo '<meta http-equiv="refresh" content="0;url=' . $url . '" />';
        App::instance()->end();
    }
    
    
    public function redirectToAdminHome() {
        $url = str_replace("/admin.php", "", App::instance()->adminBaseUrl);
        echo '<meta http-equiv="refresh" content="0;url=' . $url . '" />';
        App::instance()->end();
    }

    
    public function missingAction($actionID) {
        throw new HttpException(404, App::t('app', 'The system is unable to find the requested action "{action}".', array('{action}' => $actionID == '' ? $this->defaultAction : $actionID)));
    }

    
    public function filterAccessControl($filterChain) {
        $filter = new AccessControlFilter;
        $filter->setRules($this->accessRules());
        $filter->filter($filterChain);
    }

    
    public function beforeControllerAction($controller, $action) {
        return true;
    }

    
    public function afterControllerAction($controller, $action) {
        
    }

    
    public function render($view, $data = null, $return = false) {
        if (($viewFile = $this->getViewFile($view)) !== false) {
            $output = $this->renderFile($viewFile, $data, true);
            if ($return)
                return $output;
            else
                echo $output;
        }
        else
            throw new \Exception(App::t('app', '{controller} cannot find the requested view "{view}".', array('{controller}' => get_class($this), '{view}' => $view)));
    }

    public function renderPartial($view, $data = null, $return = false) {
        return $this->render($view, $data, $return);
    }

    
    public function getViewPath() {
        return App::instance()->getViewPath() . DIRECTORY_SEPARATOR . $this->getId();
    }

    
    public function getViewFile($viewName) {
        if (empty($viewName)) {
            return false;
        }
        if ($viewName[0] === '/') {
            $viewFile = App::instance()->getViewPath() . $viewName;
        }
        else {
            $viewFile = $this->getViewPath() . DIRECTORY_SEPARATOR . $viewName;
        }
        $extension = '.php';
        if (is_file($viewFile . $extension))
            return $viewFile . $extension;
        else {
            return false;
        }
    }

    
    public function renderFile($_viewFile_, $_data_ = null, $_return_ = false) {
                if (is_array($_data_))
            extract($_data_, EXTR_PREFIX_SAME, 'data');
        else
            $data = $_data_;
        if ($_return_) {
            ob_start();
            ob_implicit_flush(false);
            require($_viewFile_);
            return ob_get_clean();
        }
        else
            require($_viewFile_);
    }

    
    public function widget($className, $properties = array(), $captureOutput = false) {
        if ($captureOutput) {
            ob_start();
            ob_implicit_flush(false);
            try {
                $widget = $this->createWidget($className, $properties);
                $widget->run();
            }
            catch (Exception $e) {
                ob_end_clean();
                throw $e;
            }
            return ob_get_clean();
        }
        else {
            $widget = $this->createWidget($className, $properties);
            $widget->run();
            return $widget;
        }
    }

    
    public function createWidget($className, $properties = array()) {
        $widget = new $className();

        if (isset($this->widgets[$className])) {
            $properties = $properties === array() ? $this->widgets[$className] : CMap::mergeArray($this->widgets[$className], $properties);
        }
        foreach ($properties as $name => $value) {
            $widget->$name = $value;
        }
        $widget->init();
        return $widget;
    }

    
    public function beginWidget($className, $properties = array()) {
        $widget = $this->createWidget($className, $properties);
        $this->_widgetStack[] = $widget;
        return $widget;
    }

    
    public function endWidget($id = '') {
        if (($widget = array_pop($this->_widgetStack)) !== null) {
            $widget->run();
            return $widget;
        }
        else
            throw new \Exception(App::t('app', '{controller} has an extra endWidget({id}) call in its view.', array('{controller}' => get_class($this), '{id}' => $id)));
    }

}

class InlineAction {

    private $_id;
    private $_controller;

    
    public function __construct($controller, $id) {
        $this->_controller = $controller;
        $this->_id = $id;
    }

    
    public function getController() {
        return $this->_controller;
    }

    
    public function getId() {
        return $this->_id;
    }

    
    public function run() {
        $method = 'action' . $this->getId();
        $this->getController()->$method();
    }

    
    public function runWithParams($params) {
        $methodName = 'action' . $this->getId();
        $controller = $this->getController();
        $method = new \ReflectionMethod($controller, $methodName);
        if ($method->getNumberOfParameters() > 0)
            return $this->runWithParamsInternal($controller, $method, $params);

        $controller->$methodName();
        return true;
    }

    
    protected function runWithParamsInternal($object, $method, $params) {
        $ps = array();
        foreach ($method->getParameters() as $i => $param) {
            $name = $param->getName();
            if (isset($params[$name])) {
                if ($param->isArray())
                    $ps[] = is_array($params[$name]) ? $params[$name] : array($params[$name]);
                elseif (!is_array($params[$name]))
                    $ps[] = $params[$name];
                else
                    return false;
            }
            elseif ($param->isDefaultValueAvailable())
                $ps[] = $param->getDefaultValue();
            else
                return false;
        }
        $method->invokeArgs($object, $ps);
        return true;
    }

}
